'use strict';

var url = require('url');
var https = require('https');
var querystring = require('querystring');
var nconf = require.main.require('nconf');
var winston = require.main.require('winston');
var db = require.main.require('./src/database');
var User = require.main.require('./src/user');

var settings = {
	apiKey: nconf.get('youtrack_api_key'),
	categories: nconf.get('youtrack_categories'),
	project: 'QA',
	projectId: '78-12',
	projects: [],
	issuePrefix: 'https://inedo.myjetbrains.com/youtrack/issue/',
	restPrefix: 'https://inedo.myjetbrains.com/youtrack/api'
};

function init(data, callback) {
	var options = url.parse(settings.restPrefix + '/admin/projects?fields=id,name,shortName');
	options.method = 'GET';
	options.headers = {
		// do not auth!
		'Accept': 'application/json'
	};
	var req = https.request(options, async function(res) {
		var body = (await readFull(res)).toString();
		if (res.statusCode != 200) {
			return callback(new Error('YouTrack server returned ' + res.statusCode + ' ' + res.statusMessage + '\n\n' + body));
		}

		settings.projects = JSON.parse(body).map(function (p) {
			return p.shortName;
		});
		settings.projects.push('EDO');
		settings.projects.sort(function(a, b) {
			return b.length - a.length;
		});

		callback(null, data);
	});
	req.on('error', function(err) {
		callback(err);
	});
	req.end();
}

function readFull(stream) {
	var totalLength = 0;
	var buffers = [];
	return new Promise(function(resolve, reject) {
		stream.on('data', function(chunk) {
			totalLength += chunk.length;
			buffers.push(chunk);
		});
		stream.on('error', function(err) {
			reject(err);
		});
		stream.on('end', function(err) {
			resolve(Buffer.concat(buffers, totalLength));
		});
	});
}

function isSupportForum(cid) {
	return settings.categories.includes(parseInt(cid, 10));
}

function formatMessage(pid, uid, content, callback) {
	User.getUserFields(uid, ['username', 'email'], function(err, userData) {
		if (err) {
			return callback(err);
		}

		return callback(null, '<' + nconf.get('url') + '/post/' + pid + '>\n\n' +
			'From: ' + userData.username + '\n' +
			'Email: ' + userData.email + '\n' +
			'##\n' + content, userData.email);
	});
}

function onTopicPurge(data, callback) {
	callback = callback || function(err) {
		if (err) {
			winston.warn(err);
		}
	};
	db.sortedSetsRemoveRangeByScore(['inedo:youtrack:tid'], data.topic.tid, data.topic.tid, function(err) {
		callback(err, data);
	});
}

function createTicket(data, callback) {
	callback = callback || function(err) {
		if (err) {
			winston.warn(err);
		}
	};
	if (!isSupportForum(data.topic.cid)) {
		return callback(null, data);
	}

	formatMessage(data.post.pid, data.topic.uid, data.data.content, function(err, description) {
		if (err) {
			return callback(err);
		}

		var payload = JSON.stringify({
			project: { id: settings.projectId },
			summary: data.data.title,
			description: description
		});

		var options = url.parse(settings.restPrefix + '/issues?fields=idReadable');
		options.method = 'POST';
		options.headers = {
			'Authorization': 'Bearer ' + settings.apiKey,
			'Content-Type': 'application/json',
			'Content-Length': Buffer.byteLength(payload),
			'Accept': 'application/json'
		};

		var req = https.request(options, async function(res) {
			var body = (await readFull(res)).toString();
			if (res.statusCode != 200) {
				return callback(new Error('YouTrack server returned ' + res.statusCode + ' ' + res.statusMessage + '\n\n' + body));
			}
			
			if (body == null || !body.startsWith('{')) {
				return callback(new Error('YouTrack server returned an invalid response'));
			}

			var resObj = JSON.parse(body);
			var ticketNumber = resObj.idReadable;
			db.sortedSetAdd('inedo:youtrack:tid', data.topic.tid, ticketNumber, function(err) {
				callback(err, data);
			});
		});

		req.on('error', function(err) {
			callback(err);
		});

		req.write(payload);
		req.end();
	});
}

function addComment(data, callback) {
	callback = callback || function(err) {
		if (err) {
			winston.warn(err);
		}
	};
	db.getSortedSetRangeByScore('inedo:youtrack:tid', 0, 1, data.post.tid, data.post.tid, function(err, ticketNumber) {
		if (err) {
			return callback(err);
		}

		ticketNumber = ticketNumber[0];
		if (!ticketNumber) {
			return callback(null, data);
		}

		formatMessage(data.post.pid, data.post.uid, data.data.content, function (err, body, email) {
			if (err) {
				return callback(err);
			}
        
			var payload = JSON.stringify({
				query: email.toLowerCase().endsWith('@inedo.com') ? 'comment' : 'comment State Open',
				comment: body,
				issues: [{idReadable: ticketNumber}]
			});

			var options = url.parse(settings.restPrefix + 'commands');
			options.method = 'POST';
			options.headers = {
				'Authorization': 'Bearer ' + settings.apiKey,
				'Content-Type': 'application/json',
				'Content-Length': Buffer.byteLength(payload),
				'Accept': 'application/json'
			};

			var req = https.request(options, async function(res) {
				var body = (await readFull(res)).toString();
				if (res.statusCode != 200) {
					return callback(new Error('YouTrack server returned ' + res.statusCode + ' ' + res.statusMessage + '\n\n' + body));
				}

				return callback(null, data);
			});
        
			req.on('error', function(err) {
				callback(err);
			});
        
			req.write(payload);
			req.end();
		});
	});
}

function addAutoLink(parser, callback) {
	callback = callback || function(err) {
		if (err) {
			winston.warn(err);
		}
	};

	var escapeRE = parser.utils.escapeRE;
	var arrayReplaceAt = parser.utils.arrayReplaceAt;
	var boundary = '(?!-)(?:' + parser.utils.lib.ucmicro.P.source + ')|' +
		parser.utils.lib.ucmicro.Z.source + '|' +
		'[' + ' \r\n$+<=>^`|~'.split('').map(escapeRE).join('') + ']';

	parser.core.ruler.after('linkify', 'youtrack', function (state) {
		var projects = settings.projects.map(escapeRE).join('|')
		var regex = new RegExp('(^|' + boundary + ')' +
			'(' + projects + ')-([1-9][0-9]{2,})' +
			'($|' + boundary + ')', 'g');

		state.tokens.forEach(function (block) {
			if (block.type !== 'inline') {
				return;
			}

			var tokens = block.children;
			for (var i = tokens.length - 1; i >= 0; i--) {
				if (tokens[i].type !== 'text') {
					continue;
				}

				var text = tokens[i].content;
				var pos = 0;
				var nodes = [];
				regex.lastIndex = 0;
				var m, t;
				while ((m = regex.exec(text))) {
					if (m.index > 0 || m[1].length > 0) {
						t = new state.Token('text', '', 0);
						t.content = text.slice(pos, m.index + m[1].length);
						nodes.push(t);
					}

					t = new state.Token('link_open', 'a', 1);
					var prefix = settings.issuePrefix;
					if (m[2] === 'EDO') {
						prefix = 'https://my.inedo.com/tickets/view?ticketNumber=';
					}
					t.attrs = [
						['href', prefix + m[2] + '-' + m[3]],
						['class', 'inedo-link inedo-link-' + m[2].toLowerCase()]
					];
					nodes.push(t);

					t = new state.Token('text', '', 0);
					t.content = m[2] + '-' + m[3];
					nodes.push(t);

					t = new state.Token('link_close', 'a', -1);
					nodes.push(t);

					regex.lastIndex -= m[4].length;
					pos = regex.lastIndex;
				}

				if (!nodes.length) {
					continue;
				}

				if (pos < text.length) {
					t = new state.Token('text', '', 0);
					t.content = text.slice(pos);
					nodes.push(t);
				}

				block.children = tokens = arrayReplaceAt(tokens, i, nodes);
			}
		});
	});

	callback(null);
}

module.exports = {
	onLoad: init,
	onTopic: createTicket,
	onPurge: onTopicPurge,
	onReply: addComment,
	addAutoLink: addAutoLink,
};
